"""scrapli.exceptions"""


class ScrapliException(Exception):
    """Base Exception for all scrapli exceptions"""


class LibScrapliException(ScrapliException):
    """Exception raised when encountering errors loading libscrapli shared library"""


class OptionsException(ScrapliException):
    """Exception raised when encountering errors applying options"""


class AllocationException(ScrapliException):
    """Exception raised when encountering errors allocating a cli/netconf object"""


class OpenException(ScrapliException):
    """Exception raised when encountering errors opening a cli/netconf object"""


class CloseException(ScrapliException):
    """Exception raised when encountering errors closing a cli/netconf object"""


class GetResultException(ScrapliException):
    """Exception raised when encountering errors polling/fetching an operation result"""


class SubmitOperationException(ScrapliException):
    """Exception raised when encountering errors submitting an operation result"""


class NotOpenedException(ScrapliException):
    """Exception raised when attempting to call methods of cli/netconf object and ptr is None"""


class OperationException(ScrapliException):
    """Exception raised when an error is returned from an operation"""


class ParsingException(ScrapliException):
    """Exception raised when parsing (usually cli) output fails"""


class NoMessagesException(ScrapliException):
    """Exception raised when attempting to request notifications or subscriptions and none exist"""



"""scrapli.netconf_decorators"""

from collections.abc import Awaitable, Callable
from ctypes import c_uint64
from functools import update_wrapper
from typing import TYPE_CHECKING, Concatenate, ParamSpec

from scrapli.exceptions import OptionsException
from scrapli.netconf_result import Result

if TYPE_CHECKING:
    from scrapli.netconf import Netconf

P = ParamSpec("P")


def handle_operation_timeout(
    wrapped: Callable[Concatenate["Netconf", P], Result],
) -> Callable[Concatenate["Netconf", P], Result]:
    """
    Wraps a Netconf operation and sets the timeout value for the duration of the operation

    Args:
        wrapped: the operation function

    Returns:
        callable: the wrapper function

    Raises:
        N/A

    """

    def wrapper(inst: "Netconf", /, *args: P.args, **kwargs: P.kwargs) -> Result:
        """
        The operation timeout wrapper

        Args:
            inst: the Netconf instance
            args: the arguments to pass to the wrapped function
            kwargs: the keyword arguments to pass to the wrapped function

        Returns:
            Result: the result of the wrapped function

        Raises:
            OptionsException: if the operation timeout failed to set

        """
        operation_timeout_ns = kwargs.get("operation_timeout_ns")
        if operation_timeout_ns is None:
            return wrapped(inst, *args, **kwargs)

        if operation_timeout_ns == inst.session_options.operation_timeout_ns:
            return wrapped(inst, *args, **kwargs)

        if not isinstance(operation_timeout_ns, int):
            # ignore an invalid type for the timeout
            return wrapped(inst, *args, **kwargs)

        status = inst.ffi_mapping.options_mapping.session.set_operation_timeout_ns(
            inst._ptr_or_exception(),
            c_uint64(operation_timeout_ns),
        )
        if status != 0:
            raise OptionsException("failed to set session operation timeout")

        res = wrapped(inst, *args, **kwargs)

        status = inst.ffi_mapping.options_mapping.session.set_operation_timeout_ns(
            inst._ptr_or_exception(),
            c_uint64(operation_timeout_ns),
        )
        if status != 0:
            raise OptionsException("failed to set session operation timeout")

        return res

    update_wrapper(wrapper=wrapper, wrapped=wrapped)

    return wrapper


def handle_operation_timeout_async(
    wrapped: Callable[Concatenate["Netconf", P], Awaitable[Result]],
) -> Callable[Concatenate["Netconf", P], Awaitable[Result]]:
    """
    Wraps a Netconf operation and sets the timeout value for the duration of the operation

    Args:
        wrapped: the operation function

    Returns:
        callable: the wrapper function

    Raises:
        N/A

    """

    async def wrapper(inst: "Netconf", /, *args: P.args, **kwargs: P.kwargs) -> Result:
        """
        The operation timeout wrapper

        Args:
            inst: the Netconf instance
            args: the arguments to pass to the wrapped function
            kwargs: the keyword arguments to pass to the wrapped function

        Returns:
            Result: the result of the wrapped function

        Raises:
            OptionsException: if the operation timeout failed to set

        """
        operation_timeout_ns = kwargs.get("operation_timeout_ns")
        if operation_timeout_ns is None:
            return await wrapped(inst, *args, **kwargs)

        if operation_timeout_ns == inst.session_options.operation_timeout_ns:
            return await wrapped(inst, *args, **kwargs)

        if not isinstance(operation_timeout_ns, int):
            # ignore an invalid type for the timeout
            return await wrapped(inst, *args, **kwargs)

        status = inst.ffi_mapping.options_mapping.session.set_operation_timeout_ns(
            inst._ptr_or_exception(),
            c_uint64(operation_timeout_ns),
        )
        if status != 0:
            raise OptionsException("failed to set session operation timeout")

        res = await wrapped(inst, *args, **kwargs)

        status = inst.ffi_mapping.options_mapping.session.set_operation_timeout_ns(
            inst._ptr_or_exception(),
            c_uint64(operation_timeout_ns),
        )
        if status != 0:
            raise OptionsException("failed to set session operation timeout")

        return res

    update_wrapper(wrapper=wrapper, wrapped=wrapped)

    return wrapper



"""scrapli.netconf"""

from collections.abc import Callable
from ctypes import c_bool, c_char_p, c_int, c_uint, c_uint64
from dataclasses import dataclass, field
from enum import Enum
from logging import getLogger
from types import TracebackType
from typing import Any

from scrapli.auth import Options as AuthOptions
from scrapli.exceptions import (
    AllocationException,
    CloseException,
    GetResultException,
    NoMessagesException,
    NotOpenedException,
    OpenException,
    OperationException,
    OptionsException,
    SubmitOperationException,
)
from scrapli.ffi_mapping import LibScrapliMapping
from scrapli.ffi_types import (
    CancelPointer,
    DriverPointer,
    IntPointer,
    LogFuncCallback,
    OperationIdPointer,
    U64Pointer,
    ZigSlice,
    to_c_string,
)
from scrapli.helper import (
    wait_for_available_operation_result,
    wait_for_available_operation_result_async,
)
from scrapli.netconf_decorators import handle_operation_timeout, handle_operation_timeout_async
from scrapli.netconf_result import Result
from scrapli.session import Options as SessionOptions
from scrapli.transport import Options as TransportOptions


class Version(str, Enum):
    """
    Enum representing a netconf version

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    VERSION_1_0 = "1.0"
    VERSION_1_1 = "1.1"


class DatastoreType(str, Enum):
    """
    Enum representing the datastore types

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    CONVENTIONAL = "conventional"
    RUNNING = "running"
    CANDIDATE = "candidate"
    STARTUP = "startup"
    INTENDED = "intended"
    DYNAMIC = "dynamic"
    OPERATIONAL = "operational"


class FilterType(str, Enum):
    """
    Enum representing a filter type value

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    SUBTREE = "subtree"
    XPATH = "xpath"


class DefaultsType(str, Enum):
    """
    Enum representing a defaults type value

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    UNSET = "unset"
    REPORT_ALL = "report-all"
    REPORT_ALL_TAGGED = "report-all-tagged"
    TRIM = "trim"
    EXPLICIT = "explicit"


class SchemaFormat(str, Enum):
    """
    Enum representing valid schema formats

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    XSD = "xsd"
    YANG = "yang"
    YIN = "yin"
    RNG = "rng"
    RNC = "rnc"


class ConfigFilter(str, Enum):
    """
    Enum representing valid config filter values

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    UNSET = "unset"
    TRUE = "true"
    FALSE = "false"


@dataclass
class Options:
    """
    Options holds netconf related options to pass to the ffi layer.

    Args:
        error_tag: the error tag substring that identifies errors in an rpc reply
        preferred_version: preferred netconf version to use
        message_poll_interval_ns: interval in ns for message polling
        force_close: exists to enable sending "force" on close when using the context manager, this
            option causes the connection to not wait for the result of the close-session rpc. this
            can be useful if the device immediately closes the connection, not sending the "ok"
            reply.

    Returns:
        None

    Raises:
        N/A

    """

    error_tag: str | None = None
    preferred_version: Version | None = None
    message_poll_interval_ns: int | None = None
    close_force: bool = False

    _error_tag: c_char_p | None = field(init=False, default=None, repr=False)
    _preferred_version: c_char_p | None = field(init=False, default=None, repr=False)

    def apply(self, ffi_mapping: LibScrapliMapping, ptr: DriverPointer) -> None:
        """
        Applies the options to the given driver pointer.

        Should not be called directly/by users.

        Args:
            ffi_mapping: the handle to the ffi mapping singleton
            ptr: the pointer to the underlying cli or netconf object

        Returns:
            None

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        if self.error_tag is not None:
            self._error_tag = to_c_string(self.error_tag)

            status = ffi_mapping.options_mapping.netconf.set_error_tag(ptr, self._error_tag)
            if status != 0:
                raise OptionsException("failed to set netconf error tag")

        if self.preferred_version is not None:
            self._preferred_version = to_c_string(self.preferred_version)

            status = ffi_mapping.options_mapping.netconf.set_preferred_version(
                ptr, self._preferred_version
            )
            if status != 0:
                raise OptionsException("failed to set netconf preferred version")

        if self.message_poll_interval_ns is not None:
            status = ffi_mapping.options_mapping.netconf.set_message_poll_interva_ns(
                ptr, c_int(self.message_poll_interval_ns)
            )
            if status != 0:
                raise OptionsException("failed to set netconf message poll interval")

    def __repr__(self) -> str:
        """
        Magic repr method for Options object

        Args:
            N/A

        Returns:
            str: repr for Options object

        Raises:
            N/A

        """
        return (
            # it will probably be "canonical" to import Options as NetconfOptions, so we'll make
            # the repr do that too
            f"Netconf{self.__class__.__name__}("
            f"error_tag={self.error_tag!r}, "
            f"preferred_version={self.preferred_version!r} "
            f"message_poll_interval_ns={self.message_poll_interval_ns!r})"
            f"close_force={self.close_force!r})"
        )


class Netconf:
    """
    Netconf represents a netconf connection object.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(  # noqa: PLR0913
        self,
        host: str,
        *,
        logger_callback: Callable[[int, str], None] | None = None,
        port: int = 830,
        options: Options | None = None,
        auth_options: AuthOptions | None = None,
        session_options: SessionOptions | None = None,
        transport_options: TransportOptions | None = None,
        logging_uid: str | None = None,
    ) -> None:
        logger_name = f"{__name__}.{host}:{port}"
        if logging_uid is not None:
            logger_name += f":{logging_uid}"

        self.logger = getLogger(logger_name)
        self._logging_uid = logging_uid

        self.ffi_mapping = LibScrapliMapping()

        self.host = host
        self._host = to_c_string(host)

        self.logger_callback = (
            LogFuncCallback(logger_callback) if logger_callback else LogFuncCallback(0)
        )
        self._logger_callback = logger_callback

        self.port = port

        self.options = options or Options()
        self.auth_options = auth_options or AuthOptions()
        self.session_options = session_options or SessionOptions()
        self.transport_options = transport_options or TransportOptions()

        self.ptr: DriverPointer | None = None
        self._session_id: int | None = None

    def __enter__(self: "Netconf") -> "Netconf":
        """
        Enter method for context manager

        Args:
            N/A

        Returns:
            Cli: a concrete implementation of the opened Cli object

        Raises:
            ScrapliConnectionError: if an exception occurs during opening

        """
        self.open()

        return self

    def __exit__(
        self,
        exception_type: BaseException | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """
        Exit method to cleanup for context manager

        Args:
            exception_type: exception type being raised
            exception_value: message from exception being raised
            traceback: traceback from exception being raised

        Returns:
            None

        Raises:
            N/A

        """
        self.close(
            force=self.options.close_force,
        )

    async def __aenter__(self: "Netconf") -> "Netconf":
        """
        Enter method for context manager.

        Args:
            N/A

        Returns:
            Netconf: a concrete implementation of the opened Netconf object

        Raises:
            N/A

        """
        await self.open_async()

        return self

    async def __aexit__(
        self,
        exception_type: BaseException | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """
        Exit method to cleanup for context manager.

        Args:
            exception_type: exception type being raised
            exception_value: message from exception being raised
            traceback: traceback from exception being raised

        Returns:
            None

        Raises:
            N/A

        """
        await self.close_async(
            force=self.options.close_force,
        )

    def __str__(self) -> str:
        """
        Magic str method for Netconf object.

        Args:
            N/A

        Returns:
            str: str representation of Netconf

        Raises:
            N/A

        """
        return f"scrapli.Netconf {self.host}:{self.port}"

    def __repr__(self) -> str:
        """
        Magic repr method for Netconf object.

        Args:
            N/A

        Returns:
            str: repr for Netconf object

        Raises:
            N/A

        """
        return (
            f"{self.__class__.__name__}("
            f"host={self.host!r}, "
            f"port={self.port!r}, "
            f"options={self.options!r} "
            f"auth_options={self.auth_options!r} "
            f"session_options={self.auth_options!r} "
            f"transport_options={self.transport_options!r}) "
        )

    def __copy__(self, memodict: dict[Any, Any] = {}) -> "Netconf":
        # reasonably safely copy of the object... *reasonably*... basically assumes that options
        # will never be mutated during an objects lifetime, which *should* be the case. probably.
        return Netconf(
            host=self.host,
            logger_callback=self._logger_callback,
            port=self.port,
            options=self.options,
            auth_options=self.auth_options,
            session_options=self.session_options,
            transport_options=self.transport_options,
            logging_uid=self._logging_uid,
        )

    def _ptr_or_exception(self) -> DriverPointer:
        if self.ptr is None:
            raise NotOpenedException

        return self.ptr

    def _alloc(
        self,
    ) -> None:
        ptr = self.ffi_mapping.netconf_mapping.alloc(
            logger_callback=self.logger_callback,
            host=self._host,
            port=c_int(self.port),
            transport_kind=c_char_p(self.transport_options.get_transport_kind()),
        )
        if ptr == 0:  # type: ignore[comparison-overlap]
            raise AllocationException("failed to allocate netconf")

        self.ptr = ptr

        poll_fd = int(
            self.ffi_mapping.shared_mapping.get_poll_fd(
                ptr=self._ptr_or_exception(),
            )
        )
        if poll_fd == 0:
            raise AllocationException("failed to allocate netconf")

        self.poll_fd = poll_fd

    def _free(
        self,
    ) -> None:
        self.ffi_mapping.shared_mapping.free(ptr=self._ptr_or_exception())

    def _open(self) -> c_uint:
        self._alloc()

        self.options.apply(self.ffi_mapping, self._ptr_or_exception())
        self.auth_options.apply(self.ffi_mapping, self._ptr_or_exception())
        self.session_options.apply(self.ffi_mapping, self._ptr_or_exception())
        self.transport_options.apply(self.ffi_mapping, self._ptr_or_exception())

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        status = self.ffi_mapping.netconf_mapping.open(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
        )
        if status != 0:
            self._free()

            raise OpenException("failed to submit open operation")

        # cast it again, for mypy reasons
        return c_uint(operation_id.contents.value)

    def open(
        self,
    ) -> Result:
        """
        Open the netconf connection.

        Args:
            N/A

        Returns:
            None

        Raises:
            OpenException: if the operation fails

        """
        operation_id = self._open()

        return self._get_result(operation_id=operation_id)

    async def open_async(self) -> Result:
        """
        Open the netconf connection.

        Args:
            N/A

        Returns:
            None

        Raises:
            OpenException: if the operation fails

        """
        operation_id = self._open()

        return await self._get_result_async(operation_id=operation_id)

    def _close(
        self,
        force: bool,
    ) -> c_uint:
        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))
        _force = c_bool(force)

        status = self.ffi_mapping.netconf_mapping.close(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            force=_force,
        )
        if status != 0:
            raise CloseException("submitting close operation")

        return c_uint(operation_id.contents.value)

    def close(
        self,
        *,
        force: bool = False,
    ) -> Result:
        """
        Close the netconf connection.

        Args:
            force: skips sending a close-session rpc and just directly shuts down the connection

        Returns:
            None

        Raises:
            NotOpenedException: if the ptr to the netconf object is None (via _ptr_or_exception)
            CloseException: if the operation fails

        """
        operation_id = self._close(force=force)

        result = self._get_result(operation_id=operation_id)

        self._free()

        return result

    async def close_async(
        self,
        *,
        force: bool = False,
    ) -> Result:
        """
        Close the netconf connection.

        Args:
            force: skips sending a close-session rpc and just directly shuts down the connection

        Returns:
            None

        Raises:
            NotOpenedException: if the ptr to the netconf object is None (via _ptr_or_exception)
            CloseException: if the operation fails

        """
        operation_id = self._close(force=force)

        result = await self._get_result_async(operation_id=operation_id)

        self._free()

        return result

    def _get_result(
        self,
        operation_id: c_uint,
    ) -> Result:
        wait_for_available_operation_result(self.poll_fd)

        input_size = IntPointer(c_int())
        result_raw_size = IntPointer(c_int())
        result_size = IntPointer(c_int())
        rpc_warnings_size = IntPointer(c_int())
        rpc_errors_size = IntPointer(c_int())
        err_size = IntPointer(c_int())

        status = self.ffi_mapping.netconf_mapping.fetch_sizes(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            input_size=input_size,
            result_raw_size=result_raw_size,
            result_size=result_size,
            rpc_warnings_size=rpc_warnings_size,
            rpc_errors_size=rpc_errors_size,
            err_size=err_size,
        )
        if status != 0:
            raise GetResultException("wait operation failed")

        start_time = U64Pointer(c_uint64())
        end_time = U64Pointer(c_uint64())

        input_slice = ZigSlice(size=input_size.contents)
        result_raw_slice = ZigSlice(size=result_raw_size.contents)
        result_slice = ZigSlice(size=result_size.contents)

        rpc_warnings_slice = ZigSlice(size=rpc_warnings_size.contents)
        rpc_errors_slice = ZigSlice(size=rpc_errors_size.contents)
        err_slice = ZigSlice(size=err_size.contents)

        status = self.ffi_mapping.netconf_mapping.fetch(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            start_time=start_time,
            end_time=end_time,
            input_slice=input_slice,
            result_raw_slice=result_raw_slice,
            result_slice=result_slice,
            rpc_warnings_slice=rpc_warnings_slice,
            rpc_errors_slice=rpc_errors_slice,
            err_slice=err_slice,
        )
        if status != 0:
            raise GetResultException("fetch operation failed")

        err_contents = err_slice.get_decoded_contents()
        if err_contents:
            raise OperationException(err_contents)

        return Result(
            input_=input_slice.get_decoded_contents(),
            host=self.host,
            port=self.port,
            start_time=start_time.contents.value,
            end_time=end_time.contents.value,
            result_raw=result_raw_slice.get_contents(),
            result=result_slice.get_decoded_contents(),
            rpc_warnings=rpc_warnings_slice.get_decoded_contents(),
            rpc_errors=rpc_errors_slice.get_decoded_contents(),
        )

    async def _get_result_async(
        self,
        operation_id: c_uint,
    ) -> Result:
        await wait_for_available_operation_result_async(fd=self.poll_fd)

        input_size = IntPointer(c_int())
        result_raw_size = IntPointer(c_int())
        result_size = IntPointer(c_int())
        rpc_warnings_size = IntPointer(c_int())
        rpc_errors_size = IntPointer(c_int())
        err_size = IntPointer(c_int())

        status = self.ffi_mapping.netconf_mapping.fetch_sizes(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            input_size=input_size,
            result_raw_size=result_raw_size,
            result_size=result_size,
            rpc_warnings_size=rpc_warnings_size,
            rpc_errors_size=rpc_errors_size,
            err_size=err_size,
        )
        if status != 0:
            raise GetResultException("fetch operation sizes failed")

        start_time = U64Pointer(c_uint64())
        end_time = U64Pointer(c_uint64())

        input_slice = ZigSlice(size=input_size.contents)
        result_raw_slice = ZigSlice(size=result_raw_size.contents)
        result_slice = ZigSlice(size=result_size.contents)

        rpc_warnings_slice = ZigSlice(size=rpc_warnings_size.contents)
        rpc_errors_slice = ZigSlice(size=rpc_errors_size.contents)
        err_slice = ZigSlice(size=err_size.contents)

        status = self.ffi_mapping.netconf_mapping.fetch(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            start_time=start_time,
            end_time=end_time,
            input_slice=input_slice,
            result_raw_slice=result_raw_slice,
            result_slice=result_slice,
            rpc_warnings_slice=rpc_warnings_slice,
            rpc_errors_slice=rpc_errors_slice,
            err_slice=err_slice,
        )
        if status != 0:
            raise GetResultException("fetch operation failed")

        err_contents = err_slice.get_decoded_contents()
        if err_contents:
            raise OperationException(err_contents)

        return Result(
            input_=input_slice.get_decoded_contents(),
            host=self.host,
            port=self.port,
            start_time=start_time.contents.value,
            end_time=end_time.contents.value,
            result_raw=result_raw_slice.get_contents(),
            result=result_slice.get_decoded_contents(),
            rpc_warnings=rpc_warnings_slice.get_decoded_contents(),
            rpc_errors=rpc_errors_slice.get_decoded_contents(),
        )

    @property
    def session_id(self) -> int:
        """
        Get the session id of the connection.

        Args:
            N/A

        Returns:
            int: session id of the connection

        Raises:
            GetResultException: if fetching the session id fails

        """
        if self._session_id is not None:
            return self._session_id

        session_id = IntPointer(c_int())

        status = self.ffi_mapping.netconf_mapping.get_session_id(
            ptr=self._ptr_or_exception(), session_id=session_id
        )
        if status != 0:
            raise GetResultException("fetch session id failed")

        self._session_id = session_id.contents.value

        return self._session_id

    def get_subscription_id(self, payload: str) -> int:
        """
        Get the subscription id from a rpc-reply (from an establish-subscription rpc).

        Args:
            payload: the payload to find the subscription id in

        Returns:
            int: subscription id

        Raises:
            N/A

        """
        _payload = to_c_string(payload)
        subscription_id = U64Pointer(c_uint64())

        status = self.ffi_mapping.netconf_mapping.get_subscription_id(
            payload=_payload,
            subscription_id=subscription_id,
        )
        if status != 0:
            raise GetResultException("fetch subscriptiond id failed")

        return int(subscription_id.contents.value)

    def get_next_notification(
        self,
    ) -> str:
        """
        Fetch the next notification message if available.

        Args:
            N/A

        Returns:
            str: the string content of the next notification

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails
            NoMessagesException: if there are no notifications to fetch

        """
        notification_size = U64Pointer(c_uint64())

        self.ffi_mapping.netconf_mapping.get_next_notification_size(
            ptr=self._ptr_or_exception(),
            notification_size=notification_size,
        )

        if notification_size.contents == 0:
            raise NoMessagesException("no notification messages available")

        notification_slice = ZigSlice(size=notification_size.contents)

        status = self.ffi_mapping.netconf_mapping.get_next_notification(
            ptr=self._ptr_or_exception(),
            notification_slice=notification_slice,
        )
        if status != 0:
            raise SubmitOperationException("submitting getting next notification failed")

        return notification_slice.get_decoded_contents()

    def get_next_subscription(
        self,
        subscription_id: int,
    ) -> str:
        """
        Fetch the next notification message if available.

        Args:
            subscription_id: subscription id to fetch a message for

        Returns:
            str: the string content of the next notification

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails
            NoMessagesException: if there are no notifications to fetch

        """
        subscription_size = U64Pointer(c_uint64())

        self.ffi_mapping.netconf_mapping.get_next_subscription_size(
            ptr=self._ptr_or_exception(),
            subscription_id=c_uint64(subscription_id),
            subscription_size=subscription_size,
        )

        if subscription_size.contents == 0:
            raise NoMessagesException(
                f"no subscription messages available for subscription id {subscription_id}",
            )

        subscription_slice = ZigSlice(size=subscription_size.contents)

        status = self.ffi_mapping.netconf_mapping.get_next_subscription(
            ptr=self._ptr_or_exception(),
            subscription_id=c_uint64(subscription_id),
            subscription_slice=subscription_slice,
        )
        if status != 0:
            raise SubmitOperationException("submitting getting next subscription failed")

        return subscription_slice.get_decoded_contents()

    def _raw_rpc(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        payload: c_char_p,
        base_namespace_prefix: c_char_p,
        extra_namespaces: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.raw_rpc(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            payload=payload,
            base_namespace_prefix=base_namespace_prefix,
            extra_namespaces=extra_namespaces,
        )
        if status != 0:
            raise SubmitOperationException("submitting raw rpc operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def raw_rpc(
        self,
        payload: str,
        *,
        base_namespace_prefix: str = "",
        extra_namespaces: list[tuple[str, str]] | None = None,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a "raw" / user crafted rpc operation.

        Args:
            payload: the raw rpc payload
            base_namespace_prefix: prefix to use for hte base/default netconf base namespace
            extra_namespaces: optional list of pairs of prefix::namespaces. this plus the base
                namespace prefix can allow for weird cases like nxos where the base namespace must
                be prefixed and then additional namespaces indicating desired targets must be added
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _payload = to_c_string(payload)
        _base_namespace_prefix = to_c_string(base_namespace_prefix)

        if extra_namespaces is not None:
            _extra_namespaces = to_c_string(
                "__libscrapli__".join(["::".join(p) for p in extra_namespaces])
            )
        else:
            _extra_namespaces = to_c_string("")

        operation_id = self._raw_rpc(
            operation_id=operation_id,
            cancel=cancel,
            payload=_payload,
            base_namespace_prefix=_base_namespace_prefix,
            extra_namespaces=_extra_namespaces,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def raw_rpc_async(
        self,
        payload: str,
        *,
        base_namespace_prefix: str = "",
        extra_namespaces: list[tuple[str, str]] | None = None,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a "raw" / user crafted rpc operation.

        Args:
            payload: the raw rpc payload
            base_namespace_prefix: prefix to use for hte base/default netconf base namespace
            extra_namespaces: optional list of pairs of prefix::namespaces. this plus the base
                namespace prefix can allow for weird cases like nxos where the base namespace must
                be prefixed and then additional namespaces indicating desired targets must be added
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _payload = to_c_string(payload)
        _base_namespace_prefix = to_c_string(base_namespace_prefix)

        if extra_namespaces is not None:
            _extra_namespaces = to_c_string(
                "__libscrapli__".join(["::".join(p) for p in extra_namespaces])
            )
        else:
            _extra_namespaces = to_c_string("")

        operation_id = self._raw_rpc(
            operation_id=operation_id,
            cancel=cancel,
            payload=_payload,
            base_namespace_prefix=_base_namespace_prefix,
            extra_namespaces=_extra_namespaces,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _get_config(  # noqa: PLR0913,too-many-locals
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        source: c_char_p,
        filter_: c_char_p,
        filter_type: c_char_p,
        filter_namespace_prefix: c_char_p,
        filter_namespace: c_char_p,
        defaults_type: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.get_config(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            source=source,
            filter_=filter_,
            filter_type=filter_type,
            filter_namespace_prefix=filter_namespace_prefix,
            filter_namespace=filter_namespace,
            defaults_type=defaults_type,
        )
        if status != 0:
            raise SubmitOperationException("submitting get-config operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def get_config(  # noqa: PLR0913
        self,
        *,
        source: DatastoreType = DatastoreType.RUNNING,
        filter_: str = "",
        filter_type: FilterType = FilterType.SUBTREE,
        filter_namespace_prefix: str = "",
        filter_namespace: str = "",
        defaults_type: DefaultsType = DefaultsType.UNSET,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a get-config rpc operation.

        Args:
            source: source datastore to get config from
            filter_: filter to apply to the get-config (or not if empty string)
            filter_type: type of filter to apply, subtree|xpath
            filter_namespace_prefix: filter namespace prefix
            filter_namespace: filter namespace
            defaults_type: defaults type to apply to the get-config, "unset" means dont apply one
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _source = to_c_string(source)
        _filter = to_c_string(filter_)
        _filter_type = to_c_string(filter_type)
        _filter_namespace_prefix = to_c_string(filter_namespace_prefix)
        _filter_namespace = to_c_string(filter_namespace)
        _defaults_type = to_c_string(defaults_type)

        operation_id = self._get_config(
            operation_id=operation_id,
            cancel=cancel,
            source=_source,
            filter_=_filter,
            filter_type=_filter_type,
            filter_namespace_prefix=_filter_namespace_prefix,
            filter_namespace=_filter_namespace,
            defaults_type=_defaults_type,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def get_config_async(  # noqa: PLR0913
        self,
        *,
        source: DatastoreType = DatastoreType.RUNNING,
        filter_: str = "",
        filter_type: FilterType = FilterType.SUBTREE,
        filter_namespace_prefix: str = "",
        filter_namespace: str = "",
        defaults_type: DefaultsType = DefaultsType.UNSET,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a get-config rpc operation.

        Args:
            source: source datastore to get config from
            filter_: filter to apply to the get-config (or not if empty string)
            filter_type: type of filter to apply, subtree|xpath
            filter_namespace_prefix: filter namespace prefix
            filter_namespace: filter namespace
            defaults_type: defaults type to apply to the get-config, "unset" means dont apply one
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _source = to_c_string(source)
        _filter = to_c_string(filter_)
        _filter_type = to_c_string(filter_type)
        _filter_namespace_prefix = to_c_string(filter_namespace_prefix)
        _filter_namespace = to_c_string(filter_namespace)
        _defaults_type = to_c_string(defaults_type)

        operation_id = self._get_config(
            operation_id=operation_id,
            cancel=cancel,
            source=_source,
            filter_=_filter,
            filter_type=_filter_type,
            filter_namespace_prefix=_filter_namespace_prefix,
            filter_namespace=_filter_namespace,
            defaults_type=_defaults_type,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _edit_config(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        config: c_char_p,
        target: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.edit_config(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            config=config,
            target=target,
        )
        if status != 0:
            raise SubmitOperationException("submitting edit-config operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def edit_config(
        self,
        *,
        config: str = "",
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute an edit-config rpc operation.

        Args:
            config: string config payload to send
            target: target datastore as DatastoreType enum
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _config = to_c_string(config)
        _target = to_c_string(target)

        operation_id = self._edit_config(
            operation_id=operation_id,
            cancel=cancel,
            config=_config,
            target=_target,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def edit_config_async(
        self,
        *,
        config: str = "",
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute an edit-config rpc operation.

        Args:
            config: string config payload to send
            target: target datastore as DatastoreType enum
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _config = to_c_string(config)
        _target = to_c_string(target)

        operation_id = self._edit_config(
            operation_id=operation_id,
            cancel=cancel,
            config=_config,
            target=_target,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _copy_config(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
        source: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.copy_config(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            target=target,
            source=source,
        )
        if status != 0:
            raise SubmitOperationException("submitting copy-config operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def copy_config(
        self,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        source: DatastoreType = DatastoreType.STARTUP,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a copy-config rpc operation.

        Args:
            target: target to copy *to*
            source: source to copy *from*
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _target = to_c_string(target)
        _source = to_c_string(source)

        operation_id = self._copy_config(
            operation_id=operation_id,
            cancel=cancel,
            target=_target,
            source=_source,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def copy_config_async(
        self,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        source: DatastoreType = DatastoreType.STARTUP,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a copy-config rpc operation.

        Args:
            target: target to copy *to*
            source: source to copy *from*
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _target = to_c_string(target)
        _source = to_c_string(source)

        operation_id = self._copy_config(
            operation_id=operation_id,
            cancel=cancel,
            target=_target,
            source=_source,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _delete_config(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.delete_config(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            target=target,
        )
        if status != 0:
            raise SubmitOperationException("submitting delete-config operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def delete_config(
        self,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a delete-config rpc operation.

        Args:
            target: target datastore to delete
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _target = to_c_string(target)

        operation_id = self._delete_config(
            operation_id=operation_id,
            cancel=cancel,
            target=_target,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def delete_config_async(
        self,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a delete-config rpc operation.

        Args:
            target: target datastore to delete
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _target = to_c_string(target)

        operation_id = self._delete_config(
            operation_id=operation_id,
            cancel=cancel,
            target=_target,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _lock(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.lock(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            target=target,
        )
        if status != 0:
            raise SubmitOperationException("submitting lock operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def lock(
        self,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a lock rpc operation.

        Args:
            target: target datastore to lock
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _target = to_c_string(target)

        operation_id = self._lock(
            operation_id=operation_id,
            cancel=cancel,
            target=_target,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def lock_async(
        self,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a lock rpc operation.

        Args:
            target: target datastore to lock
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _target = to_c_string(target)

        operation_id = self._lock(
            operation_id=operation_id,
            cancel=cancel,
            target=_target,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _unlock(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.unlock(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            target=target,
        )
        if status != 0:
            raise SubmitOperationException("submitting unlock operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def unlock(
        self,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute an unlock rpc operation.

        Args:
            target: target datastore to unlock
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _target = to_c_string(target)

        operation_id = self._unlock(
            operation_id=operation_id,
            cancel=cancel,
            target=_target,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def unlock_async(
        self,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute an unlock rpc operation.

        Args:
            target: target datastore to unlock
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _target = to_c_string(target)

        operation_id = self._unlock(
            operation_id=operation_id,
            cancel=cancel,
            target=_target,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _get(  # noqa: PLR0913,too-many-locals
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        filter_: c_char_p,
        filter_type: c_char_p,
        filter_namespace_prefix: c_char_p,
        filter_namespace: c_char_p,
        defaults_type: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.get(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            filter_=filter_,
            filter_type=filter_type,
            filter_namespace_prefix=filter_namespace_prefix,
            filter_namespace=filter_namespace,
            defaults_type=defaults_type,
        )
        if status != 0:
            raise SubmitOperationException("submitting get operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def get(  # noqa: PLR0913
        self,
        *,
        filter_: str = "",
        filter_type: FilterType = FilterType.SUBTREE,
        filter_namespace_prefix: str = "",
        filter_namespace: str = "",
        defaults_type: DefaultsType = DefaultsType.UNSET,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a get rpc operation.

        Args:
            filter_: filter to apply to the get-config (or not if empty string)
            filter_type: type of filter to apply, subtree|xpath
            filter_namespace_prefix: filter namespace prefix
            filter_namespace: filter namespace
            defaults_type: defaults type to apply to the get-config, "unset" means dont apply one
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _filter = to_c_string(filter_)
        _filter_type = to_c_string(filter_type)
        _filter_namespace_prefix = to_c_string(filter_namespace_prefix)
        _filter_namespace = to_c_string(filter_namespace)
        _defaults_type = to_c_string(defaults_type)

        operation_id = self._get(
            operation_id=operation_id,
            cancel=cancel,
            filter_=_filter,
            filter_type=_filter_type,
            filter_namespace_prefix=_filter_namespace_prefix,
            filter_namespace=_filter_namespace,
            defaults_type=_defaults_type,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def get_async(  # noqa: PLR0913
        self,
        *,
        filter_: str = "",
        filter_type: FilterType = FilterType.SUBTREE,
        filter_namespace_prefix: str = "",
        filter_namespace: str = "",
        defaults_type: DefaultsType = DefaultsType.UNSET,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a get rpc operation.

        Args:
            filter_: filter to apply to the get-config (or not if empty string)
            filter_type: type of filter to apply, subtree|xpath
            filter_namespace_prefix: filter namespace prefix
            filter_namespace: filter namespace
            defaults_type: defaults type to apply to the get-config, "unset" means dont apply one
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _filter = to_c_string(filter_)
        _filter_type = to_c_string(filter_type)
        _filter_namespace_prefix = to_c_string(filter_namespace_prefix)
        _filter_namespace = to_c_string(filter_namespace)
        _defaults_type = to_c_string(defaults_type)

        operation_id = self._get(
            operation_id=operation_id,
            cancel=cancel,
            filter_=_filter,
            filter_type=_filter_type,
            filter_namespace_prefix=_filter_namespace_prefix,
            filter_namespace=_filter_namespace,
            defaults_type=_defaults_type,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _close_session(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.close_session(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
        )
        if status != 0:
            raise SubmitOperationException("submitting close-session operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def close_session(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a close-session rpc operation.

        Args:
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._close_session(
            operation_id=operation_id,
            cancel=cancel,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def close_session_async(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a close-session rpc operation.

        Args:
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._close_session(
            operation_id=operation_id,
            cancel=cancel,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _kill_session(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        session_id: int,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.kill_session(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            session_id=c_int(session_id),
        )
        if status != 0:
            raise SubmitOperationException("submitting kill-session operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def kill_session(
        self,
        session_id: int,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a kill-session rpc operation.

        Args:
            session_id: session id to kill
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._kill_session(
            operation_id=operation_id,
            cancel=cancel,
            session_id=session_id,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def kill_session_async(
        self,
        session_id: int,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a kill-session rpc operation.

        Args:
            session_id: session id to kill
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._kill_session(
            operation_id=operation_id,
            cancel=cancel,
            session_id=session_id,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _commit(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.commit(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
        )
        if status != 0:
            raise SubmitOperationException("submitting commit operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def commit(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a commit rpc operation.

        Args:
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._commit(
            operation_id=operation_id,
            cancel=cancel,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def commit_async(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a commit rpc operation.

        Args:
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._commit(
            operation_id=operation_id,
            cancel=cancel,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _discard(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.discard(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
        )
        if status != 0:
            raise SubmitOperationException("submitting discard operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def discard(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a discard rpc operation.

        Args:
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._discard(
            operation_id=operation_id,
            cancel=cancel,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def discard_async(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a discard rpc operation.

        Args:
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._discard(
            operation_id=operation_id,
            cancel=cancel,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _cancel_commit(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.cancel_commit(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
        )
        if status != 0:
            raise SubmitOperationException("submitting cancel-commit operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def cancel_commit(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a cancel-commit rpc operation.

        Args:
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._cancel_commit(
            operation_id=operation_id,
            cancel=cancel,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def cancel_commit_async(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a cancel-commit rpc operation.

        Args:
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._cancel_commit(
            operation_id=operation_id,
            cancel=cancel,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _validate(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        source: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.validate(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            source=source,
        )
        if status != 0:
            raise SubmitOperationException("submitting validate operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def validate(
        self,
        *,
        source: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a validate rpc operation.

        Args:
            source: datastore to validate
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _source = to_c_string(source)

        operation_id = self._validate(
            operation_id=operation_id,
            cancel=cancel,
            source=_source,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def validate_async(
        self,
        *,
        source: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a validate rpc operation.

        Args:
            source: datastore to validate
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _source = to_c_string(source)

        operation_id = self._validate(
            operation_id=operation_id,
            cancel=cancel,
            source=_source,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _get_schema(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        identifier: c_char_p,
        version: c_char_p,
        format_: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.get_schema(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            identifier=identifier,
            version=version,
            format_=format_,
        )
        if status != 0:
            raise SubmitOperationException("submitting get-schema operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def get_schema(
        self,
        identifier: str,
        *,
        version: str = "",
        format_: SchemaFormat = SchemaFormat.YANG,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a get-schema rpc operation.

        Args:
            identifier: schema identifier to get
            version: optional schema version to request
            format_: schema format to apply
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _identifier = to_c_string(identifier)
        _version = to_c_string(version)
        _format = to_c_string(format_)

        operation_id = self._get_schema(
            operation_id=operation_id,
            cancel=cancel,
            identifier=_identifier,
            version=_version,
            format_=_format,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def get_schema_async(
        self,
        identifier: str,
        *,
        version: str = "",
        format_: SchemaFormat = SchemaFormat.YANG,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a get-schema rpc operation.

        Args:
            identifier: schema identifier to get
            version: optional schema version to request
            format_: schema format to apply
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _identifier = to_c_string(identifier)
        _version = to_c_string(version)
        _format = to_c_string(format_)

        operation_id = self._get_schema(
            operation_id=operation_id,
            cancel=cancel,
            identifier=_identifier,
            version=_version,
            format_=_format,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _get_data(  # noqa: PLR0913,too-many-locals
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        source: c_char_p,
        filter_: c_char_p,
        filter_type: c_char_p,
        filter_namespace_prefix: c_char_p,
        filter_namespace: c_char_p,
        config_filter: c_char_p,
        origin_filters: c_char_p,
        max_depth: int,
        with_origin: bool,
        defaults_type: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.get_data(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            source=source,
            filter_=filter_,
            filter_type=filter_type,
            filter_namespace_prefix=filter_namespace_prefix,
            filter_namespace=filter_namespace,
            config_filter=config_filter,
            origin_filters=origin_filters,
            max_depth=c_int(max_depth),
            with_origin=c_bool(with_origin),
            defaults_type=defaults_type,
        )
        if status != 0:
            raise SubmitOperationException("submitting copy-config operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def get_data(  # noqa: PLR0913
        self,
        *,
        source: DatastoreType = DatastoreType.RUNNING,
        filter_: str = "",
        filter_type: FilterType = FilterType.SUBTREE,
        filter_namespace_prefix: str = "",
        filter_namespace: str = "",
        config_filter: ConfigFilter = ConfigFilter.UNSET,
        origin_filters: str = "",
        max_depth: int = 0,
        with_origin: bool = False,
        defaults_type: DefaultsType = DefaultsType.UNSET,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a get-data rpc operation.

        Args:
            source: source datastore to get data from
            filter_: filter to apply to the get-config (or not if empty string)
            filter_type: type of filter to apply, subtree|xpath
            filter_namespace_prefix: filter namespace prefix
            filter_namespace: filter namespace
            config_filter: config filter true/false, or unset to leave up to the server
            origin_filters: fully formed origin filter xml payload to embed
            max_depth: max depth of data requested
            with_origin: include origin data
            defaults_type: defaults type to apply to the get-config, "unset" means dont apply one
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _source = to_c_string(source)
        _filter = to_c_string(filter_)
        _filter_type = to_c_string(filter_type)
        _filter_namespace_prefix = to_c_string(filter_namespace_prefix)
        _filter_namespace = to_c_string(filter_namespace)
        _config_filter = to_c_string(config_filter)
        _origin_filters = to_c_string(origin_filters)
        _defaults_type = to_c_string(defaults_type)

        operation_id = self._get_data(
            operation_id=operation_id,
            cancel=cancel,
            source=_source,
            filter_=_filter,
            filter_type=_filter_type,
            filter_namespace_prefix=_filter_namespace_prefix,
            filter_namespace=_filter_namespace,
            config_filter=_config_filter,
            origin_filters=_origin_filters,
            max_depth=max_depth,
            with_origin=with_origin,
            defaults_type=_defaults_type,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def get_data_async(  # noqa: PLR0913
        self,
        *,
        source: DatastoreType = DatastoreType.RUNNING,
        filter_: str = "",
        filter_type: FilterType = FilterType.SUBTREE,
        filter_namespace_prefix: str = "",
        filter_namespace: str = "",
        config_filter: ConfigFilter = ConfigFilter.UNSET,
        origin_filters: str = "",
        max_depth: int = 0,
        with_origin: bool = False,
        defaults_type: DefaultsType = DefaultsType.UNSET,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute a get-data rpc operation.

        Args:
            source: source datastore to get data from
            filter_: filter to apply to the get-config (or not if empty string)
            filter_type: type of filter to apply, subtree|xpath
            filter_namespace_prefix: filter namespace prefix
            filter_namespace: filter namespace
            config_filter: config filter true/false, or unset to leave up to the server
            origin_filters: fully formed origin filter xml payload to embed
            max_depth: max depth of data requested
            with_origin: include origin data
            defaults_type: defaults type to apply to the get-config, "unset" means dont apply one
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _source = to_c_string(source)
        _filter = to_c_string(filter_)
        _filter_type = to_c_string(filter_type)
        _filter_namespace_prefix = to_c_string(filter_namespace_prefix)
        _filter_namespace = to_c_string(filter_namespace)
        _config_filter = to_c_string(config_filter)
        _origin_filters = to_c_string(origin_filters)
        _defaults_type = to_c_string(defaults_type)

        operation_id = self._get_data(
            operation_id=operation_id,
            cancel=cancel,
            source=_source,
            filter_=_filter,
            filter_type=_filter_type,
            filter_namespace_prefix=_filter_namespace_prefix,
            filter_namespace=_filter_namespace,
            config_filter=_config_filter,
            origin_filters=_origin_filters,
            max_depth=max_depth,
            with_origin=with_origin,
            defaults_type=_defaults_type,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _edit_data(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        content: c_char_p,
        target: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.edit_data(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            content=content,
            target=target,
        )
        if status != 0:
            raise SubmitOperationException("submitting copy-config operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def edit_data(
        self,
        content: str,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute an edit-data rpc operation.

        Args:
            content: full payload content to send
            target: datastore to target
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _content = to_c_string(content)
        _target = to_c_string(target)

        operation_id = self._edit_data(
            operation_id=operation_id,
            cancel=cancel,
            content=_content,
            target=_target,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def edit_data_async(
        self,
        content: str,
        *,
        target: DatastoreType = DatastoreType.RUNNING,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute an edit-data rpc operation.

        Args:
            content: full payload content to send
            target: datastore to target
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _content = to_c_string(content)
        _target = to_c_string(target)

        operation_id = self._edit_data(
            operation_id=operation_id,
            cancel=cancel,
            content=_content,
            target=_target,
        )

        return await self._get_result_async(operation_id=operation_id)

    def _action(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        action: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.netconf_mapping.action(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            action=action,
        )
        if status != 0:
            raise SubmitOperationException("submitting action operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def action(
        self,
        action: str,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute an action rpc operation.

        Args:
            action: action to execute
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _action = to_c_string(action)

        operation_id = self._action(
            operation_id=operation_id,
            cancel=cancel,
            action=_action,
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def action_async(
        self,
        action: str,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Execute an action rpc operation.

        Args:
            action: action to execute
            operation_timeout_ns: optional timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _action = to_c_string(action)

        operation_id = self._action(
            operation_id=operation_id,
            cancel=cancel,
            action=_action,
        )

        return await self._get_result_async(operation_id=operation_id)



"""scrapli.ffi_mapping"""

from collections.abc import Callable
from ctypes import (
    CDLL,
    c_bool,
    c_char_p,
    c_int,
    c_uint8,
    c_uint32,
)

from scrapli.ffi import get_libscrapli_path
from scrapli.ffi_mapping_cli import LibScrapliCliMapping
from scrapli.ffi_mapping_netconf import LibScrapliNetconfMapping
from scrapli.ffi_mapping_options import LibScrapliOptionsMapping
from scrapli.ffi_types import (
    DriverPointer,
    IntPointer,
    StringPointer,
)


class LibScrapliSharedMapping:
    """
    Mapping to libscrapli shared (between cli/netconf) object functions mapping.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self._get_poll_fd: Callable[
            [
                DriverPointer,
            ],
            c_uint32,
        ] = lib.ls_shared_get_poll_fd
        lib.ls_shared_get_poll_fd.argtypes = [
            DriverPointer,
        ]
        lib.ls_cli_alloc.restype = c_uint32

        self._free: Callable[
            [
                DriverPointer,
            ],
            int,
        ] = lib.ls_shared_free
        lib.ls_shared_free.argtypes = [
            DriverPointer,
        ]
        lib.ls_shared_free.restype = None

        self._read: Callable[
            [
                DriverPointer,
                StringPointer,
                IntPointer,
            ],
            int,
        ] = lib.ls_shared_read_session
        lib.ls_shared_read_session.argtypes = [
            DriverPointer,
            StringPointer,
            IntPointer,
        ]
        lib.ls_shared_read_session.restype = c_int

        self._write: Callable[
            [
                DriverPointer,
                c_char_p,
                c_bool,
            ],
            int,
        ] = lib.ls_shared_write_session
        lib.ls_shared_write_session.argtypes = [
            DriverPointer,
            c_char_p,
            c_bool,
        ]
        lib.ls_shared_write_session.restype = c_uint8

    def get_poll_fd(self, ptr: DriverPointer) -> c_uint32:
        """
        Get the operation poll fd from the driver at ptr.

        Should (generally) not be called directly/by users.

        Args:
            ptr: the ptr to the libscrapli cli/netconf object.

        Returns:
            c_uint32: the poll fd for the driver.

        Raises:
            N/A

        """
        return self._get_poll_fd(ptr)

    def free(self, ptr: DriverPointer) -> int:
        """
        Free the driver at ptr.

        Should (generally) not be called directly/by users.

        Args:
            ptr: the ptr to the libscrapli cli/netconf object.

        Returns:
            c_uint8: return code, non-zero value indicates an error.

        Raises:
            N/A

        """
        return self._free(ptr)

    def read(self, ptr: DriverPointer, buf: StringPointer, read_size: IntPointer) -> int:
        """
        Read from the session of driver at ptr.

        Should (generally) not be called directly/by users.

        Args:
            ptr: the ptr to the libscrapli cli/netconf object.
            buf: buffer to fill during the read operation.
            read_size: c_int pointer that is filled with number of bytes written into buf.

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._read(ptr, buf, read_size)

    def write(self, ptr: DriverPointer, buf: c_char_p, redacted: c_bool) -> int:
        """
        Write to the session of driver at ptr.

        Should (generally) not be called directly/by users.

        Args:
            ptr: the ptr to the libscrapli cli/netconf object.
            buf: buffer contents to write during the write operation..
            redacted: bool indicated if the write contents should be redacted from logs.

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._write(ptr, buf, redacted)


class LibScrapliMapping:
    """
    Single mapping to libscrapli shared library exported functions.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    _instance = None

    def __new__(cls) -> "LibScrapliMapping":
        """Returns the singleton instance of the ffi mapping"""
        if not cls._instance:
            cls._instance = super().__new__(cls)

        return cls._instance

    def __init__(self) -> None:
        self.lib = CDLL(get_libscrapli_path())
        self.shared_mapping = LibScrapliSharedMapping(self.lib)
        self.cli_mapping = LibScrapliCliMapping(self.lib)
        self.netconf_mapping = LibScrapliNetconfMapping(self.lib)
        self.options_mapping = LibScrapliOptionsMapping(self.lib)



"""scrapli.transport"""

from ctypes import c_char_p, c_int
from dataclasses import dataclass, field
from enum import Enum

from scrapli.exceptions import OptionsException
from scrapli.ffi_mapping import LibScrapliMapping
from scrapli.ffi_types import (
    DriverPointer,
    to_c_string,
)


class TransportKind(str, Enum):
    """
    Enum representing the transport kind

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    BIN = "bin"
    SSH2 = "ssh2"
    TELNET = "telnet"
    TEST = "test_"


@dataclass
class BinOptions:
    """
    Options holds bin transport options to pass to the ffi layer.

    Args:
        bin: path to binary to use for the transport
        extra_open_args: extra args to pass to the binary
        override_open_args: override all "normal" scrapli args with these args
        ssh_config_path: path to ssh config file
        known_hosts_path: path to known hosts file
        enable_strict_key: enable ssh strict key checking
        term_height: set the terminal height
        term_width: set the terminal width

    Returns:
        None

    Raises:
        N/A

    """

    bin: str | None = None
    extra_open_args: list[str] | None = None
    override_open_args: list[str] | None = None
    ssh_config_path: str | None = None
    known_hosts_path: str | None = None
    enable_strict_key: bool | None = None
    term_height: int | None = None
    term_width: int | None = None

    _bin: c_char_p | None = field(init=False, default=None, repr=False)
    _extra_open_args: c_char_p | None = field(init=False, default=None, repr=False)
    _override_open_args: c_char_p | None = field(init=False, default=None, repr=False)
    _ssh_config_path: c_char_p | None = field(init=False, default=None, repr=False)
    _known_hosts_path: c_char_p | None = field(init=False, default=None, repr=False)

    def apply(  # noqa: C901, PLR0912
        self, ffi_mapping: LibScrapliMapping, ptr: DriverPointer
    ) -> None:
        """
        Applies the options to the given driver pointer.

        Should not be called directly/by users.

        Args:
            ffi_mapping: the handle to the ffi mapping singleton
            ptr: the pointer to the underlying cli or netconf object

        Returns:
            None

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        if self.bin is not None:
            self._bin = to_c_string(self.bin)

            status = ffi_mapping.options_mapping.transport_bin.set_bin(ptr, self._bin)
            if status != 0:
                raise OptionsException("failed to set bin transport bin")

        if self.extra_open_args is not None:
            self._extra_open_args = to_c_string(" ".join(self.extra_open_args))

            status = ffi_mapping.options_mapping.transport_bin.set_extra_open_args(
                ptr, self._extra_open_args
            )
            if status != 0:
                raise OptionsException("failed to set bin transport extra open args")

        if self.override_open_args is not None:
            self._override_open_args = to_c_string(" ".join(self.override_open_args))

            status = ffi_mapping.options_mapping.transport_bin.set_override_open_args(
                ptr, self._override_open_args
            )
            if status != 0:
                raise OptionsException("failed to set bin transport override open args")

        if self.ssh_config_path is not None:
            self._ssh_config_path = to_c_string(self.ssh_config_path)

            status = ffi_mapping.options_mapping.transport_bin.set_ssh_config_path(
                ptr, self._ssh_config_path
            )
            if status != 0:
                raise OptionsException("failed to set bin transport ssh config path")

        if self.known_hosts_path is not None:
            self._known_hosts_path = to_c_string(self.known_hosts_path)

            status = ffi_mapping.options_mapping.transport_bin.set_known_hosts_path(
                ptr, self._known_hosts_path
            )
            if status != 0:
                raise OptionsException("failed to set bin transport ssh known hosts path")

        if self.enable_strict_key is not None:
            status = ffi_mapping.options_mapping.transport_bin.set_enable_strict_key(ptr)
            if status != 0:
                raise OptionsException("failed to set bin transport enable strict key")

        if self.term_height is not None:
            status = ffi_mapping.options_mapping.transport_bin.set_term_height(
                ptr, c_int(self.term_height)
            )
            if status != 0:
                raise OptionsException("failed to set bin transport term height")

        if self.term_width is not None:
            status = ffi_mapping.options_mapping.transport_bin.set_term_width(
                ptr, c_int(self.term_width)
            )
            if status != 0:
                raise OptionsException("failed to set bin transport term width")


@dataclass
class Ssh2Options:
    """
    Options holds ssh2 transport options to pass to the ffi layer.

    Args:
        libssh2_trace: enable libssh2 tracing

    Returns:
        None

    Raises:
        N/A

    """

    libssh2_trace: bool | None = None

    def apply(self, ffi_mapping: LibScrapliMapping, ptr: DriverPointer) -> None:
        """
        Applies the options to the given driver pointer.

        Should not be called directly/by users.

        Args:
            ffi_mapping: the handle to the ffi mapping singleton
            ptr: the pointer to the underlying cli or netconf object

        Returns:
            None

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        if self.libssh2_trace is not None:
            status = ffi_mapping.options_mapping.transport_ssh2.set_libssh2_trace(ptr)
            if status != 0:
                raise OptionsException("failed to set ssh2 transport trace")


@dataclass
class TelnetOptions:
    """
    Options holds telnet transport options to pass to the ffi layer.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def apply(self, ffi_mapping: LibScrapliMapping, ptr: DriverPointer) -> None:
        """
        Applies the options to the given driver pointer.

        Should not be called directly/by users.

        Args:
            ffi_mapping: the handle to the ffi mapping singleton
            ptr: the pointer to the underlying cli or netconf object

        Returns:
            None

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        _, _ = ffi_mapping, ptr


@dataclass
class TestOptions:
    """
    Options holds test transport options to pass to the ffi layer.

    Args:
        f: the file to load/read for the transport

    Returns:
        None

    Raises:
        N/A

    """

    f: str | None = None

    _f: c_char_p | None = field(init=False, default=None, repr=False)

    def apply(self, ffi_mapping: LibScrapliMapping, ptr: DriverPointer) -> None:
        """
        Applies the options to the given driver pointer.

        Should not be called directly/by users.

        Args:
            ffi_mapping: the handle to the ffi mapping singleton
            ptr: the pointer to the underlying cli or netconf object

        Returns:
            None

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        if self.f is not None:
            self._f = to_c_string(self.f)

            status = ffi_mapping.options_mapping.transport_test.set_f(ptr, self._f)
            if status != 0:
                raise OptionsException("failed to set test transport f")


@dataclass
class Options:
    """
    Options holds transport options to pass to the ffi layer.

    Only one flavor of transport options can be set at a time. The active option determines the
    transport to be used for the connection.

    Args:
        bin: the bin transport options
        ssh2: the ssh2 transport options
        telnet: the telnet transport options
        test: the test transport options

    Returns:
        None

    Raises:
        N/A

    """

    bin: BinOptions | None = None
    ssh2: Ssh2Options | None = None
    telnet: TelnetOptions | None = None
    test: TestOptions | None = None

    _transport_kind: bytes | None = field(init=False, default=None, repr=False)

    def __post_init__(self) -> None:
        _set_fields = [f for f in [self.bin, self.ssh2, self.telnet, self.test] if f is not None]

        if len(_set_fields) == 0:
            self.bin = BinOptions()

            return

        if len(_set_fields) == 1:
            return

        raise OptionsException("more than one transport set")

    def apply(self, ffi_mapping: LibScrapliMapping, ptr: DriverPointer) -> None:
        """
        Applies the options to the given driver pointer.

        Should not be called directly/by users.

        Args:
            ffi_mapping: the handle to the ffi mapping singleton
            ptr: the pointer to the underlying cli or netconf object

        Returns:
            None

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        if self.bin:
            self.bin.apply(ffi_mapping, ptr)
        elif self.ssh2:
            self.ssh2.apply(ffi_mapping, ptr)
        elif self.telnet:
            self.telnet.apply(ffi_mapping, ptr)
        elif self.test:
            self.test.apply(ffi_mapping, ptr)

    def get_transport_kind(self) -> bytes:
        """
        Returns the encoded name of the selected transport kind

        Should not be called directly/by users.

        Args:
            N/A

        Returns:
            bytes: the encoded name of the selected transport kind

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        if self.bin:
            self._transport_kind = TransportKind.BIN.encode(encoding="utf-8")
        elif self.ssh2:
            self._transport_kind = TransportKind.SSH2.encode(encoding="utf-8")
        elif self.telnet:
            self._transport_kind = TransportKind.TELNET.encode(encoding="utf-8")
        else:
            self._transport_kind = TransportKind.TEST.encode(encoding="utf-8")

        return self._transport_kind

    def __repr__(self) -> str:
        """
        Magic repr method for Options object

        Args:
            N/A

        Returns:
            str: repr for Options object

        Raises:
            N/A

        """
        if self.bin:
            return (
                # it will probably be "canonical" to import Options as SessionOptions, so we'll make
                # the repr do that too
                f"Transport{self.__class__.__name__}("
                f"bin={self.bin!r}, "
            )

        if self.ssh2:
            return f"Transport{self.__class__.__name__}(" f"ssh2={self.ssh2!r}, "

        if self.telnet:
            return f"Transport{self.__class__.__name__}(" f"telnet={self.telnet!r}, "

        return f"Transport{self.__class__.__name__}(" f"test={self.test!r}, "



"""scrapli"""

from scrapli.auth import LookupKeyValue
from scrapli.auth import Options as AuthOptions
from scrapli.cli import Cli
from scrapli.netconf import Netconf
from scrapli.netconf import Options as NetconfOptions
from scrapli.session import Options as SessionOptions
from scrapli.transport import BinOptions as TransportBinOptions
from scrapli.transport import Options as TransportOptions
from scrapli.transport import Ssh2Options as TransportSsh2Options
from scrapli.transport import TelnetOptions as TransportTelnetOptions
from scrapli.transport import TestOptions as TransportTestOptions

__version__ = "2.0.0-alpha.3"
__definitions_version__ = "471f12e"

__all__ = (
    "AuthOptions",
    "Cli",
    "LookupKeyValue",
    "Netconf",
    "NetconfOptions",
    "SessionOptions",
    "TransportBinOptions",
    "TransportOptions",
    "TransportSsh2Options",
    "TransportTelnetOptions",
    "TransportTestOptions",
)



"""scrapli.ffi_mapping_cli"""

from collections.abc import Callable
from ctypes import (
    CDLL,
    c_bool,
    c_char_p,
    c_int,
    c_uint8,
)

from _ctypes import POINTER

from scrapli.ffi_types import (
    CancelPointer,
    DriverPointer,
    IntPointer,
    LogFuncCallback,
    OperationId,
    OperationIdPointer,
    U64Pointer,
    ZigSlice,
    ZigSlicePointer,
    ZigU64Slice,
)


class LibScrapliCliMapping:
    """
    Mapping to libscrapli cli object functions mapping.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self._alloc: Callable[
            [
                c_char_p,
                LogFuncCallback,
                c_char_p,
                c_int,
                c_char_p,
            ],
            DriverPointer,
        ] = lib.ls_cli_alloc
        lib.ls_cli_alloc.argtypes = [
            c_char_p,
            LogFuncCallback,
            c_char_p,
            c_int,
            c_char_p,
        ]
        lib.ls_cli_alloc.restype = DriverPointer

        self._get_ntc_templates_platform: Callable[
            [
                DriverPointer,
                ZigSlicePointer,
            ],
            int,
        ] = lib.ls_cli_get_ntc_templates_platform
        lib.ls_cli_get_ntc_templates_platform.argtypes = [
            DriverPointer,
            POINTER(ZigSlice),
        ]
        lib.ls_cli_get_ntc_templates_platform.restype = c_uint8

        self._get_genie_platform: Callable[
            [
                DriverPointer,
                ZigSlicePointer,
            ],
            int,
        ] = lib.ls_cli_get_genie_platform
        lib.ls_cli_get_genie_platform.argtypes = [
            DriverPointer,
            POINTER(ZigSlice),
        ]
        lib.ls_cli_get_genie_platform.restype = c_uint8

        self._open: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
            ],
            int,
        ] = lib.ls_cli_open
        lib.ls_cli_open.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
        ]
        lib.ls_cli_open.restype = c_uint8

        self._close: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
            ],
            int,
        ] = lib.ls_cli_close
        lib.ls_cli_close.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
        ]
        lib.ls_cli_close.restype = c_uint8

        self._fetch_sizes: Callable[
            [
                DriverPointer,
                OperationId,
                IntPointer,
                IntPointer,
                IntPointer,
                IntPointer,
                IntPointer,
                IntPointer,
            ],
            int,
        ] = lib.ls_cli_fetch_operation_sizes
        lib.ls_cli_fetch_operation_sizes.argtypes = [
            DriverPointer,
            OperationId,
            IntPointer,
            IntPointer,
            IntPointer,
            IntPointer,
            IntPointer,
            IntPointer,
        ]
        lib.ls_cli_fetch_operation_sizes.restype = c_uint8

        self._fetch: Callable[
            [
                DriverPointer,
                OperationId,
                IntPointer,
                ZigU64Slice,
                ZigSlicePointer,
                ZigSlicePointer,
                ZigSlicePointer,
                ZigSlicePointer,
                ZigSlicePointer,
            ],
            int,
        ] = lib.ls_cli_fetch_operation
        lib.ls_cli_fetch_operation.argtypes = [
            DriverPointer,
            OperationId,
            U64Pointer,
            POINTER(ZigU64Slice),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
        ]
        lib.ls_cli_fetch_operation.restype = c_uint8

        self._enter_mode: Callable[
            [DriverPointer, OperationIdPointer, CancelPointer, c_char_p], int
        ] = lib.ls_cli_enter_mode
        lib.ls_cli_enter_mode.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
        ]
        lib.ls_cli_enter_mode.restype = c_uint8

        self._get_prompt: Callable[[DriverPointer, OperationIdPointer, CancelPointer], int] = (
            lib.ls_cli_get_prompt
        )
        lib.ls_cli_get_prompt.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
        ]
        lib.ls_cli_get_prompt.restype = c_uint8

        self._send_input: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
                c_char_p,
                c_bool,
                c_bool,
            ],
            int,
        ] = lib.ls_cli_send_input
        lib.ls_cli_send_input.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
            c_char_p,
            c_bool,
            c_bool,
        ]
        lib.ls_cli_send_input.restype = c_uint8

        self._send_prompted_input: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_bool,
                c_bool,
            ],
            int,
        ] = lib.ls_cli_send_prompted_input
        lib.ls_cli_send_prompted_input.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_bool,
            c_bool,
        ]
        lib.ls_cli_send_prompted_input.restype = c_uint8

    def alloc(
        self,
        *,
        definition_string: c_char_p,
        logger_callback: LogFuncCallback,
        host: c_char_p,
        port: c_int,
        transport_kind: c_char_p,
    ) -> DriverPointer:
        """
        Allocate a cli object.

        Should (generally) not be called directly/by users.

        Args:
            definition_string: yaml definition string
            logger_callback: pointer to logger callback function
            host: host to connect to
            port: port at which to connect
            transport_kind: transport kind to use

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._alloc(definition_string, logger_callback, host, port, transport_kind)

    def get_ntc_templates_platform(
        self,
        *,
        ptr: DriverPointer,
        ntc_templates_platform: ZigSlicePointer,
    ) -> int:
        """
        Writes the ntc templates platform into the given slice pointer.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the cli object
            ntc_templates_platform: slice to write the ntc templates platform into

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_ntc_templates_platform(
            ptr,
            ntc_templates_platform,
        )

    def get_genie_platform(
        self,
        *,
        ptr: DriverPointer,
        genie_platform: ZigSlicePointer,
    ) -> int:
        """
        Writes the (cisco/pyats) genie platform into the given slice pointer.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the cli object
            genie_platform: slice to write the ntc templates platform into

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_genie_platform(
            ptr,
            genie_platform,
        )

    def open(
        self, ptr: DriverPointer, operation_id: OperationIdPointer, cancel: CancelPointer
    ) -> int:
        """
        Open the driver at ptr.

        Should (generally) not be called directly/by users.

        Args:
            ptr: the ptr to the libscrapli cli object.
            operation_id: c_int pointer that is filled with the operation id to poll for completion.
            cancel: bool pointer that can be set to true to cancel the operation.

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._open(ptr, operation_id, cancel)

    def close(
        self,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> int:
        """
        Close the driver at ptr.

        Should (generally) not be called directly/by users.

        Args:
            ptr: the ptr to the libscrapli cli object.
            operation_id: c_int pointer that is filled with the operation id to poll for completion.
            cancel: bool pointer that can be set to true to cancel the operation.

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._close(
            ptr,
            operation_id,
            cancel,
        )

    def fetch_sizes(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationId,
        operation_count: IntPointer,
        inputs_size: IntPointer,
        results_raw_size: IntPointer,
        results_size: IntPointer,
        results_failed_indicator_size: IntPointer,
        err_size: IntPointer,
    ) -> int:
        """
        Fetch the sizes of a cli operation's results.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the cli object
            operation_id: operation id of which to poll
            operation_count: int pointer to fill with the count of operations
            inputs_size: int pointer to fill with the operation's input size
            results_raw_size: int pointer to fill with the operation's result raw size
            results_size:  int pointer to fill with the operation's result size
            results_failed_indicator_size: int pointer to fill with the operation's failed indicator
                size
            err_size: int pointer to fill with the operation's error size

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._fetch_sizes(
            ptr,
            operation_id,
            operation_count,
            inputs_size,
            results_raw_size,
            results_size,
            results_failed_indicator_size,
            err_size,
        )

    def fetch(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationId,
        start_time: U64Pointer,
        splits: ZigU64Slice,
        inputs_slice: ZigSlicePointer,
        results_raw_slice: ZigSlicePointer,
        results_slice: ZigSlicePointer,
        results_failed_indicator_slice: ZigSlicePointer,
        err_slice: ZigSlicePointer,
    ) -> int:
        """
        Fetch the result of a cli operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the cli object
            operation_id: operation id of which to poll
            start_time: int pointer to fill with the operation's start time
            splits: slice of u64 timestamps -- the end time (in unix ns) for each operation
            inputs_slice: pre allocated slice to fill with the operations input
            results_raw_slice: pre allocated slice to fill with the operations result raw
            results_slice: pre allocated slice to fill with the operations result
            results_failed_indicator_slice: pre allocated slice to fill with the operations failed
                indicator
            err_slice: pre allocated slice to fill with the operations error

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._fetch(
            ptr,
            operation_id,
            start_time,
            splits,
            inputs_slice,
            results_raw_slice,
            results_slice,
            results_failed_indicator_slice,
            err_slice,
        )

    def enter_mode(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        requested_mode: c_char_p,
    ) -> int:
        """
        Enter the given mode for the cli object.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the cli object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            requested_mode: string name of the mode to enter

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._enter_mode(ptr, operation_id, cancel, requested_mode)

    def get_prompt(
        self, *, ptr: DriverPointer, operation_id: OperationIdPointer, cancel: CancelPointer
    ) -> int:
        """
        Get the current prompt for the cli object.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the cli object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_prompt(ptr, operation_id, cancel)

    def send_input(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        input_: c_char_p,
        requested_mode: c_char_p,
        input_handling: c_char_p,
        retain_input: c_bool,
        retain_trailing_prompt: c_bool,
    ) -> int:
        """
        Send some input to the cli object.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the cli object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            input_: the input to send
            requested_mode: string name of the mode to send the input in
            input_handling: string mapping to input handling enum that governs how the input is
                handled
            retain_input: boolean indicating whether to retain the input after entering
            retain_trailing_prompt: boolean indicating whether to retain the trailing

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._send_input(
            ptr,
            operation_id,
            cancel,
            input_,
            requested_mode,
            input_handling,
            retain_input,
            retain_trailing_prompt,
        )

    def send_prompted_input(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        input_: c_char_p,
        prompt: c_char_p,
        prompt_pattern: c_char_p,
        response: c_char_p,
        abort_input: c_char_p,
        requested_mode: c_char_p,
        input_handling: c_char_p,
        hidden_response: c_bool,
        retain_trailing_prompt: c_bool,
    ) -> int:
        """
        Send some prompted input to the cli object.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the cli object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            input_: the input to send
            prompt: the prompt to expect
            prompt_pattern: the prompt pattern to expect
            response: the response to write when the prompt has been seen
            abort_input: the input to send to abort the "prompted input" operation if an error
                is encountered
            requested_mode: string name of the mode to send the input in
            input_handling: string mapping to input handling enum that governs how the input is
                handled
            hidden_response: bool indicated if the response we write will be "hidden" on the device
            retain_trailing_prompt: boolean indicating whether to retain the trailing

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._send_prompted_input(
            ptr,
            operation_id,
            cancel,
            input_,
            prompt,
            prompt_pattern,
            response,
            abort_input,
            requested_mode,
            input_handling,
            hidden_response,
            retain_trailing_prompt,
        )



"""scrapli.ffi_mapping_cli"""

from collections.abc import Callable
from ctypes import (
    CDLL,
    c_bool,
    c_char_p,
    c_int,
    c_uint8,
    c_uint64,
)

from _ctypes import POINTER

from scrapli.ffi_types import (
    CancelPointer,
    DriverPointer,
    IntPointer,
    LogFuncCallback,
    OperationId,
    OperationIdPointer,
    U64Pointer,
    ZigSlice,
    ZigSlicePointer,
)


class LibScrapliNetconfMapping:
    """
    Mapping to libscrapli netconf object functions mapping.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:  # noqa: PLR0915
        self._alloc: Callable[
            [
                LogFuncCallback,
                c_char_p,
                c_int,
                c_char_p,
            ],
            DriverPointer,
        ] = lib.ls_netconf_alloc
        lib.ls_netconf_alloc.argtypes = [
            LogFuncCallback,
            c_char_p,
            c_int,
            c_char_p,
        ]
        lib.ls_netconf_alloc.restype = DriverPointer

        self._open: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
            ],
            int,
        ] = lib.ls_netconf_open
        lib.ls_netconf_open.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
        ]
        lib.ls_netconf_open.restype = c_uint8

        self._close: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_bool,
            ],
            int,
        ] = lib.ls_netconf_close
        lib.ls_netconf_close.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_bool,
        ]
        lib.ls_netconf_close.restype = c_uint8

        self._fetch_sizes: Callable[
            [
                DriverPointer,
                OperationId,
                IntPointer,
                IntPointer,
                IntPointer,
                IntPointer,
                IntPointer,
                IntPointer,
            ],
            int,
        ] = lib.ls_netconf_fetch_operation_sizes
        lib.ls_netconf_fetch_operation_sizes.argtypes = [
            DriverPointer,
            OperationId,
            IntPointer,
            IntPointer,
            IntPointer,
            IntPointer,
            IntPointer,
            IntPointer,
        ]
        lib.ls_netconf_fetch_operation_sizes.restype = c_uint8

        self._fetch: Callable[
            [
                DriverPointer,
                OperationId,
                U64Pointer,
                U64Pointer,
                ZigSlicePointer,
                ZigSlicePointer,
                ZigSlicePointer,
                ZigSlicePointer,
                ZigSlicePointer,
                ZigSlicePointer,
            ],
            int,
        ] = lib.ls_netconf_fetch_operation
        lib.ls_netconf_fetch_operation.argtypes = [
            DriverPointer,
            OperationId,
            POINTER(c_uint64),
            POINTER(c_uint64),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
            POINTER(ZigSlice),
        ]
        lib.ls_netconf_fetch_operation.restype = c_uint8

        self._get_session_id: Callable[
            [
                DriverPointer,
                IntPointer,
            ],
            int,
        ] = lib.ls_netconf_get_session_id
        lib.ls_netconf_get_session_id.argtypes = [
            DriverPointer,
            IntPointer,
        ]
        lib.ls_netconf_get_session_id.restype = c_uint8

        self._get_subscription_id: Callable[
            [
                c_char_p,
                U64Pointer,
            ],
            int,
        ] = lib.ls_netconf_get_subscription_id
        lib.ls_netconf_get_subscription_id.argtypes = [
            c_char_p,
            POINTER(c_uint64),
        ]
        lib.ls_netconf_get_subscription_id.restype = c_uint8

        self._get_next_notification_size: Callable[
            [
                DriverPointer,
                U64Pointer,
            ],
            int,
        ] = lib.ls_netconf_next_notification_message_size
        lib.ls_netconf_next_notification_message_size.argtypes = [
            DriverPointer,
            POINTER(c_uint64),
        ]
        lib.ls_netconf_next_notification_message_size.restype = c_uint8

        self._get_next_notification: Callable[
            [
                DriverPointer,
                ZigSlicePointer,
            ],
            int,
        ] = lib.ls_netconf_next_notification_message
        lib.ls_netconf_next_notification_message.argtypes = [
            DriverPointer,
            POINTER(ZigSlice),
        ]
        lib.ls_netconf_next_notification_message.restype = c_uint8

        self._get_next_subscription_size: Callable[
            [
                DriverPointer,
                c_uint64,
                U64Pointer,
            ],
            int,
        ] = lib.ls_netconf_next_subscription_message_size
        lib.ls_netconf_next_subscription_message_size.argtypes = [
            DriverPointer,
            c_uint64,
            POINTER(c_uint64),
        ]
        lib.ls_netconf_next_subscription_message_size.restype = c_uint8

        self._get_next_subscription: Callable[
            [
                DriverPointer,
                c_uint64,
                ZigSlicePointer,
            ],
            int,
        ] = lib.ls_netconf_next_subscription_message
        lib.ls_netconf_next_subscription_message.argtypes = [
            DriverPointer,
            c_uint64,
            POINTER(ZigSlice),
        ]
        lib.ls_netconf_next_subscription_message.restype = c_uint8

        self._raw_rpc: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_raw_rpc
        lib.ls_netconf_raw_rpc.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
            c_char_p,
        ]
        lib.ls_netconf_raw_rpc.restype = c_uint8

        self._get_config: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_get_config
        lib.ls_netconf_get_config.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
        ]
        lib.ls_netconf_get_config.restype = c_uint8

        self._edit_config: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_edit_config
        lib.ls_netconf_edit_config.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
        ]
        lib.ls_netconf_edit_config.restype = c_uint8

        self._copy_config: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_copy_config
        lib.ls_netconf_copy_config.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
        ]
        lib.ls_netconf_copy_config.restype = c_uint8

        self._delete_config: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_delete_config
        lib.ls_netconf_delete_config.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
        ]
        lib.ls_netconf_delete_config.restype = c_uint8

        self._lock: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_lock
        lib.ls_netconf_lock.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
        ]
        lib.ls_netconf_lock.restype = c_uint8

        self._unlock: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_unlock
        lib.ls_netconf_unlock.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
        ]
        lib.ls_netconf_unlock.restype = c_uint8

        self._get: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_get
        lib.ls_netconf_get.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
        ]
        lib.ls_netconf_get.restype = c_uint8

        self._close_session: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
            ],
            int,
        ] = lib.ls_netconf_close_session
        lib.ls_netconf_close_session.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
        ]
        lib.ls_netconf_close_session.restype = c_uint8

        self._kill_session: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_int,
            ],
            int,
        ] = lib.ls_netconf_kill_session
        lib.ls_netconf_kill_session.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_int,
        ]
        lib.ls_netconf_kill_session.restype = c_uint8

        self._commit: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
            ],
            int,
        ] = lib.ls_netconf_commit
        lib.ls_netconf_commit.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
        ]
        lib.ls_netconf_commit.restype = c_uint8

        self._discard: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
            ],
            int,
        ] = lib.ls_netconf_discard
        lib.ls_netconf_discard.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
        ]
        lib.ls_netconf_discard.restype = c_uint8

        self._cancel_commit: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
            ],
            int,
        ] = lib.ls_netconf_cancel_commit
        lib.ls_netconf_cancel_commit.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
        ]
        lib.ls_netconf_cancel_commit.restype = c_uint8

        self._validate: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_validate
        lib.ls_netconf_validate.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
        ]
        lib.ls_netconf_validate.restype = c_uint8

        self._get_schema: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_get_schema
        lib.ls_netconf_get_schema.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
            c_char_p,
        ]
        lib.ls_netconf_get_schema.restype = c_uint8

        self._get_data: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_char_p,
                c_int,
                c_bool,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_get_data
        lib.ls_netconf_get_data.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_char_p,
            c_int,
            c_bool,
            c_char_p,
        ]
        lib.ls_netconf_get_data.restype = c_uint8

        self._edit_data: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_edit_data
        lib.ls_netconf_edit_data.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
            c_char_p,
        ]
        lib.ls_netconf_edit_data.restype = c_uint8

        self._action: Callable[
            [
                DriverPointer,
                OperationIdPointer,
                CancelPointer,
                c_char_p,
            ],
            int,
        ] = lib.ls_netconf_action
        lib.ls_netconf_action.argtypes = [
            DriverPointer,
            OperationIdPointer,
            CancelPointer,
            c_char_p,
        ]
        lib.ls_netconf_action.restype = c_uint8

    def alloc(
        self,
        *,
        logger_callback: LogFuncCallback,
        host: c_char_p,
        port: c_int,
        transport_kind: c_char_p,
    ) -> DriverPointer:
        """
        Allocate a Netconf object.

        Should (generally) not be called directly/by users.

        Args:
            logger_callback: pointer to logger callback function
            host: host to connect to
            port: port at which to connect
            transport_kind: transport kind to use

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._alloc(logger_callback, host, port, transport_kind)

    def open(
        self, ptr: DriverPointer, operation_id: OperationIdPointer, cancel: CancelPointer
    ) -> int:
        """
        Open the driver at ptr.

        Should (generally) not be called directly/by users.

        Args:
            ptr: the ptr to the libscrapli netconf object.
            operation_id: c_int pointer that is filled with the operation id to poll for completion.
            cancel: bool pointer that can be set to true to cancel the operation.

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._open(ptr, operation_id, cancel)

    def close(
        self,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        force: c_bool,
    ) -> int:
        """
        Close the driver at ptr.

        Should (generally) not be called directly/by users.

        Args:
            ptr: the ptr to the libscrapli netconf object.
            operation_id: c_int pointer that is filled with the operation id to poll for completion.
            cancel: bool pointer that can be set to true to cancel the operation.
            force: bool indicating if the connection should skip sending close-session rpc or not

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._close(
            ptr,
            operation_id,
            cancel,
            force,
        )

    def fetch_sizes(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationId,
        input_size: IntPointer,
        result_raw_size: IntPointer,
        result_size: IntPointer,
        rpc_warnings_size: IntPointer,
        rpc_errors_size: IntPointer,
        err_size: IntPointer,
    ) -> int:
        """
        Fetch the sizes of a netconf operation's results.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: operation id of which to poll
            input_size: int pointer to fill with the operation's input size
            result_raw_size: int pointer to fill with the operation's result raw size
            result_size:  int pointer to fill with the operation's result size
            rpc_warnings_size: int pointer to fill with the size of any rpc warning string
            rpc_errors_size: int pointer to fill with the size of any rpc error string
            err_size: int pointer to fill with the operation's error size

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._fetch_sizes(
            ptr,
            operation_id,
            input_size,
            result_raw_size,
            result_size,
            rpc_warnings_size,
            rpc_errors_size,
            err_size,
        )

    def fetch(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationId,
        start_time: U64Pointer,
        end_time: U64Pointer,
        input_slice: ZigSlicePointer,
        result_raw_slice: ZigSlicePointer,
        result_slice: ZigSlicePointer,
        rpc_warnings_slice: ZigSlicePointer,
        rpc_errors_slice: ZigSlicePointer,
        err_slice: ZigSlicePointer,
    ) -> int:
        """
        Fetch the result of a cli operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: operation id of which to poll
            start_time: int pointer to fill with the operation's start time
            end_time: int pointer to fill with the operation's end time
            input_slice: pre allocated slice to fill with the operations input
            result_raw_slice: pre allocated slice to fill with the operations result raw
            result_slice: pre allocated slice to fill with the operations result
            rpc_warnings_slice: pre allocated slice to fill with the rpc warnings string
            rpc_errors_slice: pre allocated slice to fill with the rpc errors string
            err_slice: pre allocated slice to fill with the operations error

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._fetch(
            ptr,
            operation_id,
            start_time,
            end_time,
            input_slice,
            result_raw_slice,
            result_slice,
            rpc_warnings_slice,
            rpc_errors_slice,
            err_slice,
        )

    def get_session_id(
        self,
        *,
        ptr: DriverPointer,
        session_id: IntPointer,
    ) -> int:
        """
        Get the session id of the Netconf object/session.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            session_id: int pointer to fill with the session id

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_session_id(
            ptr,
            session_id,
        )

    def get_subscription_id(
        self,
        *,
        payload: c_char_p,
        subscription_id: U64Pointer,
    ) -> int:
        """
        Get the subscription id from the given subscription response payload.

        Should (generally) not be called directly/by users.

        Args:
            payload: payload to get the subscription id from
            subscription_id: int pointer to fill with the subscription id

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_subscription_id(
            payload,
            subscription_id,
        )

    def get_next_notification_size(
        self,
        *,
        ptr: DriverPointer,
        notification_size: U64Pointer,
    ) -> int:
        """
        Get the size of the next notification message.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            notification_size: int pointer to fill with the session id

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_next_notification_size(
            ptr,
            notification_size,
        )

    def get_next_notification(
        self,
        *,
        ptr: DriverPointer,
        notification_slice: ZigSlicePointer,
    ) -> int:
        """
        Get the next notification message.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            notification_slice: pre populted slice to fill with the notification

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_next_notification(
            ptr,
            notification_slice,
        )

    def get_next_subscription_size(
        self,
        *,
        ptr: DriverPointer,
        subscription_id: c_uint64,
        subscription_size: U64Pointer,
    ) -> int:
        """
        Get the size of the next subscription message for the given id.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            subscription_id: subscription id to fetch a message for
            subscription_size: int pointer to fill with the session id

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_next_subscription_size(
            ptr,
            subscription_id,
            subscription_size,
        )

    def get_next_subscription(
        self,
        *,
        ptr: DriverPointer,
        subscription_id: c_uint64,
        subscription_slice: ZigSlicePointer,
    ) -> int:
        """
        Get the next subscription message for the given id.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            subscription_id: subscription id to fetch a message for
            subscription_slice: pre populted slice to fill with the notification

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_next_subscription(
            ptr,
            subscription_id,
            subscription_slice,
        )

    def raw_rpc(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        payload: c_char_p,
        base_namespace_prefix: c_char_p,
        extra_namespaces: c_char_p,
    ) -> int:
        """
        Execute a "raw" / user defined rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            payload: the payload to write into the outer rpc element
            base_namespace_prefix: prefix to use for hte base/default netconf base namespace
            extra_namespaces: extra namespace::prefix pairs (using "::" as split there), and
                split by "__libscrapli__" for additional pairs. this plus the base namespace
                prefix can allow for weird cases like nxos where the base namespace must be
                prefixed and then additional namespaces indicating desired targets must be added

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._raw_rpc(
            ptr,
            operation_id,
            cancel,
            payload,
            base_namespace_prefix,
            extra_namespaces,
        )

    def get_config(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        source: c_char_p,
        filter_: c_char_p,
        filter_type: c_char_p,
        filter_namespace_prefix: c_char_p,
        filter_namespace: c_char_p,
        defaults_type: c_char_p,
    ) -> int:
        """
        Execute a get-config rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            source: source data store to get config from
            filter_: filter to apply
            filter_type: filter type (subtree|xpath)
            filter_namespace_prefix: optional prefix for filter namespace
            filter_namespace: optional namespace for filter
            defaults_type: defaults type setting

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_config(
            ptr,
            operation_id,
            cancel,
            source,
            filter_,
            filter_type,
            filter_namespace_prefix,
            filter_namespace,
            defaults_type,
        )

    def edit_config(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        config: c_char_p,
        target: c_char_p,
    ) -> int:
        """
        Execute an edit-config rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            config: the config to send
            target: the target datastore

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._edit_config(
            ptr,
            operation_id,
            cancel,
            config,
            target,
        )

    def copy_config(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
        source: c_char_p,
    ) -> int:
        """
        Execute a copy-config rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            target: the target/destination datastore to copy to
            source: the source datastore to copy from

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._copy_config(
            ptr,
            operation_id,
            cancel,
            target,
            source,
        )

    def delete_config(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
    ) -> int:
        """
        Execute a delete-config rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            target: the target/destination datastore to delete

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._delete_config(
            ptr,
            operation_id,
            cancel,
            target,
        )

    def lock(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
    ) -> int:
        """
        Execute a lock rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            target: the target/destination datastore to lock

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._lock(
            ptr,
            operation_id,
            cancel,
            target,
        )

    def unlock(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
    ) -> int:
        """
        Execute an unlock rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            target: the target/destination datastore to lock

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._unlock(
            ptr,
            operation_id,
            cancel,
            target,
        )

    def get(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        filter_: c_char_p,
        filter_type: c_char_p,
        filter_namespace_prefix: c_char_p,
        filter_namespace: c_char_p,
        defaults_type: c_char_p,
    ) -> int:
        """
        Execute a get rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            filter_: filter to apply
            filter_type: filter type (subtree|xpath)
            filter_namespace_prefix: optional prefix for filter namespace
            filter_namespace: optional namespace for filter
            defaults_type: defaults type setting

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get(
            ptr,
            operation_id,
            cancel,
            filter_,
            filter_type,
            filter_namespace_prefix,
            filter_namespace,
            defaults_type,
        )

    def close_session(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> int:
        """
        Execute a close-config rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._close_session(
            ptr,
            operation_id,
            cancel,
        )

    def kill_session(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        session_id: c_int,
    ) -> int:
        """
        Execute a close-config rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            session_id: the session id to kill

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._kill_session(
            ptr,
            operation_id,
            cancel,
            session_id,
        )

    def commit(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> int:
        """
        Execute a commit rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._commit(
            ptr,
            operation_id,
            cancel,
        )

    def discard(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> int:
        """
        Execute a discard rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._discard(
            ptr,
            operation_id,
            cancel,
        )

    def cancel_commit(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> int:
        """
        Execute a cancel-commit rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._cancel_commit(
            ptr,
            operation_id,
            cancel,
        )

    def validate(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        source: c_char_p,
    ) -> int:
        """
        Execute a validate rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            source: datastore to validate

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._validate(
            ptr,
            operation_id,
            cancel,
            source,
        )

    def get_schema(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        identifier: c_char_p,
        version: c_char_p,
        format_: c_char_p,
    ) -> int:
        """
        Execute a get-schema rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            identifier: schema identifier to get
            version: optional schema version to request
            format_: schema format to apply

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_schema(
            ptr,
            operation_id,
            cancel,
            identifier,
            version,
            format_,
        )

    def get_data(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        source: c_char_p,
        filter_: c_char_p,
        filter_type: c_char_p,
        filter_namespace_prefix: c_char_p,
        filter_namespace: c_char_p,
        config_filter: c_char_p,
        origin_filters: c_char_p,
        max_depth: c_int,
        with_origin: c_bool,
        defaults_type: c_char_p,
    ) -> int:
        """
        Execute a get-data rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            source: source datastore to get data from
            filter_: filter to apply to the get-config (or not if empty string)
            filter_type: type of filter to apply, subtree|xpath
            filter_namespace_prefix: filter namespace prefix
            filter_namespace: filter namespace
            config_filter: config filter true/false, or unset to leave up to the server
            origin_filters: fully formed origin filter xml payload to embed
            max_depth: max depth of data requested
            with_origin: include origin data
            defaults_type: defaults type to apply to the get-config, "unset" means dont apply one

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._get_data(
            ptr,
            operation_id,
            cancel,
            source,
            filter_,
            filter_type,
            filter_namespace_prefix,
            filter_namespace,
            config_filter,
            origin_filters,
            max_depth,
            with_origin,
            defaults_type,
        )

    def edit_data(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        target: c_char_p,
        content: c_char_p,
    ) -> int:
        """
        Execute a edit-data rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            target: datastore to target
            content: full payload content to send

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._edit_data(
            ptr,
            operation_id,
            cancel,
            target,
            content,
        )

    def action(
        self,
        *,
        ptr: DriverPointer,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        action: c_char_p,
    ) -> int:
        """
        Execute an action rpc operation.

        Should (generally) not be called directly/by users.

        Args:
            ptr: ptr to the netconf object
            operation_id: int pointer to fill with the id of the submitted operation
            cancel: bool pointer that can be set to true to cancel the operation
            action: the action to send

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._action(
            ptr,
            operation_id,
            cancel,
            action,
        )
